@@ -66,7 +66,8 @@ struct SnStartupSequence
   Time timestamp;
 
   char *binary_name;
-  char *icon_name;  
+  char *icon_name;
+  char *application_id;
 
   unsigned int completed : 1;
   unsigned int canceled : 1;
@@ -269,7 +270,8 @@ sn_startup_sequence_unref (SnStartupSequence *sequence)
       sn_free (sequence->wmclass);
       sn_free (sequence->binary_name);
       sn_free (sequence->icon_name);
-      
+      sn_free (sequence->application_id);
+
       sn_display_unref (sequence->display);
       sn_free (sequence);
     }
@@ -337,6 +339,12 @@ sn_startup_sequence_get_icon_name (SnStartupSequence *sequence)
   return sequence->icon_name;
 }
 
+const char*
+sn_startup_sequence_get_application_id (SnStartupSequence *sequence)
+{
+  return sequence->application_id;
+}
+
 int
 sn_startup_sequence_get_screen (SnStartupSequence *sequence)
 {
@@ -412,8 +420,8 @@ sn_startup_sequence_complete (SnStartupSequence *sequence)
 
   sn_internal_broadcast_xmessage (sequence->display,
                                   sequence->screen,
-                                  "_NET_STARTUP_INFO",
-                                  "_NET_STARTUP_INFO_BEGIN",
+                                  sn_internal_get_net_startup_info_atom(sequence->display),
+                                  sn_internal_get_net_startup_info_begin_atom(sequence->display),
                                   message);
 
   sn_free (message);
@@ -596,8 +604,7 @@ dispatch_monitor_event (SnDisplay      *display,
 }
 
 sn_bool_t
-sn_internal_monitor_process_event (SnDisplay *display,
-                                   XEvent    *xevent)
+sn_internal_monitor_process_event (SnDisplay *display)
 {
   sn_bool_t retval;
 
@@ -625,8 +632,8 @@ find_sequence_by_id_foreach (void *value,
   FindSequenceByIdData *fsd = data;
   
   if (strcmp (sequence->id, fsd->id) == 0 &&
-      sn_display_get_x_display (sequence->display) ==
-      sn_display_get_x_display (fsd->display))
+      sn_internal_display_get_id (sequence->display) ==
+      sn_internal_display_get_id (fsd->display))
     {
       fsd->found = sequence;
       return FALSE;
@@ -783,7 +790,7 @@ xmessage_func (SnDisplay  *display,
                 {
                   int n;
                   n = atoi (values[i]);
-                  if (n >= 0 && n < ScreenCount (sn_display_get_x_display (sequence->display)))
+                  if (n >= 0 && n < sn_internal_display_get_screen_number (sequence->display))
                     {
                       sequence->screen = n;
                       changed = TRUE;
@@ -806,10 +813,18 @@ xmessage_func (SnDisplay  *display,
                   changed = TRUE;
                 }
             }
+          else if (strcmp (names[i], "APPLICATION_ID") == 0)
+            {
+              if (sequence->application_id == NULL)
+                {
+                  sequence->application_id = sn_internal_strdup (values[i]);
+                  changed = TRUE;
+                }
+            }
           else if (strcmp (names[i], "DESKTOP") == 0)
             {
               int workspace;
-              
+
               workspace = sn_internal_string_to_ulong (values[i]);
 
               sequence->workspace = workspace;
